#include <utility>



#ifndef PORYGONLANG_BOUNDSTATEMENT_HPP
#define PORYGONLANG_BOUNDSTATEMENT_HPP
#include <utility>
#include <vector>
#include "../BoundExpressions/BoundExpression.hpp"
#include "../BoundVariables/BoundVariableKey.hpp"


using namespace std;

namespace Porygon::Binder {
    enum class BoundStatementKind {
        Bad,
        Script,
        Block,
        Expression,
        Assignment,
        IndexAssignment,
        FunctionDeclaration,
        Return,
        Conditional,
    };

    class BoundStatement {
    public:
        virtual const BoundStatementKind GetKind() const = 0;

        virtual ~BoundStatement() = default;
    };

    class BoundBadStatement : public BoundStatement {
    public:
        const BoundStatementKind GetKind() const final {
            return BoundStatementKind::Bad;
        }
    };

    class BoundBlockStatement : public BoundStatement {
        const vector<BoundStatement *> _statements;
    public:
        explicit BoundBlockStatement(vector<BoundStatement *> statements)
                : _statements(std::move(statements)) {
        }

        ~BoundBlockStatement() override {
            for (auto s : _statements) {
                delete s;
            }
        }

        const BoundStatementKind GetKind() const override {
            return BoundStatementKind::Block;
        }

        const vector<BoundStatement *> *GetStatements() const {
            return &_statements;
        }
    };

    class BoundScriptStatement : public BoundBlockStatement {
        const int _localVariableCount;
    public:
        explicit BoundScriptStatement(vector<BoundStatement *> statements, int localVariableCount)
                : BoundBlockStatement(std::move(statements)),
                  _localVariableCount(localVariableCount) {
        }

        const BoundStatementKind GetKind() const final {
            return BoundStatementKind::Script;
        }

        const int GetLocalVariableCount() const {
            return _localVariableCount;
        }
    };

    class BoundExpressionStatement : public BoundStatement {
        const BoundExpression *_expression;
    public:
        explicit BoundExpressionStatement(BoundExpression *expression)
                : _expression(expression) {
            _expression = expression;
        }

        ~BoundExpressionStatement() final {
            delete _expression;
        }

        const BoundStatementKind GetKind() const final {
            return BoundStatementKind::Expression;
        }

        const BoundExpression *GetExpression() const {
            return _expression;
        }
    };

    class BoundAssignmentStatement : public BoundStatement {
        const BoundVariableKey *_key;
        const BoundExpression *_expression;
    public:
        BoundAssignmentStatement(BoundVariableKey *key, BoundExpression *expression)
                : _key(key), _expression(expression) {
        }

        ~BoundAssignmentStatement() final {
            delete _key;
            delete _expression;
        }

        const BoundStatementKind GetKind() const final {
            return BoundStatementKind::Assignment;
        }

        const BoundVariableKey *GetKey() const {
            return _key;
        }

        const BoundExpression *GetExpression() const {
            return _expression;
        }
    };

    class BoundIndexAssignmentStatement : public BoundStatement {
        const BoundExpression *_indexExpression;
        const BoundExpression *_valueExpression;
    public:
        BoundIndexAssignmentStatement(const BoundExpression *index, BoundExpression *value)
                : _indexExpression(index), _valueExpression(value) {
        }

        ~BoundIndexAssignmentStatement() final {
            delete _indexExpression;
            delete _valueExpression;
        }

        const BoundStatementKind GetKind() const final {
            return BoundStatementKind::IndexAssignment;
        }

        const BoundExpression *GetIndexExpression() const {
            return _indexExpression;
        }

        const BoundExpression *GetValueExpression() const {
            return _valueExpression;
        }
    };

    class BoundReturnStatement : public BoundStatement {
        const BoundExpression *_expression;
    public:
        explicit BoundReturnStatement(BoundExpression *expression)
                : _expression(expression) {
        }

        ~BoundReturnStatement() final {
            delete _expression;
        }

        const BoundStatementKind GetKind() const final {
            return BoundStatementKind::Return;
        }

        const BoundExpression *GetExpression() const {
            return _expression;
        }
    };

    class BoundConditionalStatement : public BoundStatement {
        const BoundExpression *_condition;
        const BoundStatement *_block;
        const BoundStatement *_elseStatement;
    public:
        explicit BoundConditionalStatement(BoundExpression *condition, BoundStatement *block, BoundStatement *next)
                : _condition(condition), _block(block), _elseStatement(next) {
        }

        ~BoundConditionalStatement() final {
            delete _condition;
            delete _block;
            delete _elseStatement;
        }

        const BoundStatementKind GetKind() const final {
            return BoundStatementKind::Conditional;
        }

        const BoundExpression *GetCondition() const {
            return _condition;
        }

        const BoundStatement *GetBlock() const {
            return _block;
        }

        const BoundStatement *GetElseStatement() const {
            return _elseStatement;
        }
    };
}


#include "BoundFunctionDeclarationStatement.hpp"

#endif //PORYGONLANG_BOUNDSTATEMENT_HPP
