#include <utility>

#include <utility>

#ifndef PORYGONLANG_PARSEDSTATEMENT_HPP
#define PORYGONLANG_PARSEDSTATEMENT_HPP

#include <utility>
#include <vector>
#include <string>

#include "../ParsedExpressions/ParsedExpression.hpp"
#include "../../Utilities/HashedString.hpp"
#include "../TypedVariableIdentifier.hpp"

namespace Porygon::Parser {
    enum class ParsedStatementKind {
        Bad,
        Script,
        Block,
        Expression,
        Assignment,
        IndexAssignment,
        FunctionDeclaration,
        Return,
        Conditional
    };

    class ParsedStatement {
        const unsigned int _start;
        const unsigned int _length;
    public:
        ParsedStatement(unsigned int start, unsigned int length)
                : _start(start), _length(length) {
        }

        virtual ~ParsedStatement() = default;

        virtual const ParsedStatementKind GetKind() const = 0;

        const unsigned int GetStartPosition() const {
            return _start;
        }

        const unsigned int GetLength() const {
            return _length;
        }

        const unsigned int GetEndPosition() const {
            return _start + _length - 1;
        }
    };

    class ParsedBadStatement : public ParsedStatement {
    public:
        ParsedBadStatement(unsigned int start, unsigned int length) : ParsedStatement(start, length) {};

        const ParsedStatementKind GetKind() const final {
            return ParsedStatementKind::Bad;
        }
    };

    class ParsedBlockStatement : public ParsedStatement {
        const std::vector<const ParsedStatement *> _statements;
    public:
        explicit ParsedBlockStatement(std::vector<const ParsedStatement *> statements)
                : ParsedStatement(statements.front()->GetStartPosition(),
                                  statements.back()->GetEndPosition() - statements.front()->GetStartPosition()),
                  _statements(statements) {}

        ParsedBlockStatement(std::vector<const ParsedStatement *> statements, unsigned int start) : ParsedStatement(
                start, 0),
                                                                                                    _statements(
                                                                                                            std::move(
                                                                                                                    statements)) {
        }


        ~ParsedBlockStatement() override {
            for (auto s: _statements) {
                delete s;
            }
        }

        const ParsedStatementKind GetKind() const override {
            return ParsedStatementKind::Block;
        }

        const std::vector<const ParsedStatement *> *GetStatements() const {
            return &_statements;
        }
    };

    class ParsedScriptStatement : public ParsedBlockStatement {
    public:
        explicit ParsedScriptStatement(vector<const ParsedStatement *> statements) : ParsedBlockStatement(
                move(statements)) {}

        const ParsedStatementKind GetKind() const final {
            return ParsedStatementKind::Script;
        }
    };

    class ParsedExpressionStatement : public ParsedStatement {
        const ParsedExpression *_expression;
    public:
        explicit ParsedExpressionStatement(ParsedExpression *expression)
                : ParsedStatement(expression->GetStartPosition(), expression->GetLength()),
                  _expression(expression) {
        }

        ~ParsedExpressionStatement() override {
            delete _expression;
        }

        const ParsedStatementKind GetKind() const final {
            return ParsedStatementKind::Expression;
        }

        const ParsedExpression *GetExpression() const {
            return _expression;
        }
    };

    class ParsedFunctionDeclarationStatement : public ParsedStatement {
        const HashedString _identifier;
        const vector<TypedVariableIdentifier *> _parameters;
        const ParsedBlockStatement *_block;
    public:
        ParsedFunctionDeclarationStatement(HashedString identifier, vector<TypedVariableIdentifier *> parameters,
                                           ParsedBlockStatement *block,
                                           unsigned int start, unsigned int length)
                : ParsedStatement(start, length), _identifier(identifier), _parameters(std::move(parameters)),
                  _block(block) {};

        ~ParsedFunctionDeclarationStatement() override {
            for (auto v : _parameters) {
                delete v;
            }
            delete _block;
        }

        const ParsedStatementKind GetKind() const final {
            return ParsedStatementKind::FunctionDeclaration;
        }

        const HashedString GetIdentifier() const {
            return _identifier;
        }

        const vector<TypedVariableIdentifier *> *GetParameters() const {
            return &_parameters;
        }

        const ParsedBlockStatement *GetBlock() const {
            return _block;
        }
    };

    class ParsedAssignmentStatement : public ParsedStatement {
        const bool _local;
        const HashedString _identifier;
        const ParsedExpression *_expression;
    public:
        ParsedAssignmentStatement(bool local, const HashedString identifier, ParsedExpression *expression,
                                  unsigned int start, unsigned int length)
                : ParsedStatement(start, length), _local(local), _identifier(identifier), _expression(expression) {
        }

        ~ParsedAssignmentStatement() final {
            delete _expression;
        }

        const ParsedStatementKind GetKind() const final {
            return ParsedStatementKind::Assignment;
        }

        const bool IsLocal() const {
            return _local;
        }

        const HashedString GetIdentifier() const {
            return _identifier;
        }

        const ParsedExpression *GetExpression() const {
            return _expression;
        }
    };

    class ParsedIndexAssignmentStatement : public ParsedStatement {
        const ParsedExpression *_indexExpression;
        const ParsedExpression *_valueExpression;
    public:
        ParsedIndexAssignmentStatement(ParsedExpression *indexExpression, ParsedExpression *valueExpression,
                                       unsigned int start,
                                       unsigned int length)
                : ParsedStatement(start, length),
                  _indexExpression(indexExpression), _valueExpression(valueExpression) {}

        ~ParsedIndexAssignmentStatement() final {
            delete _indexExpression;
            delete _valueExpression;
        }

        const ParsedStatementKind GetKind() const final {
            return ParsedStatementKind::IndexAssignment;
        }

        const ParsedExpression *GetIndexExpression() const {
            return _indexExpression;
        }

        const ParsedExpression *GetValueExpression() const {
            return _valueExpression;
        }
    };

    class ParsedReturnStatement : public ParsedStatement {
        const ParsedExpression *_expression;
    public:
        ParsedReturnStatement(ParsedExpression *expression, unsigned int start, unsigned int length)
                : ParsedStatement(start, length),
                  _expression(expression) {
        }

        ~ParsedReturnStatement() final {
            delete _expression;
        }

        const ParsedStatementKind GetKind() const final {
            return ParsedStatementKind::Return;
        }

        const ParsedExpression *GetExpression() const {
            return _expression;
        }
    };

    class ParsedConditionalStatement : public ParsedStatement {
        const ParsedExpression *_condition;
        const ParsedStatement *_block;
        // This can be either else if or else
        const ParsedStatement *_elseStatement;
    public:
        ParsedConditionalStatement(ParsedExpression *condition, ParsedStatement *block, unsigned int start,
                                   unsigned int length)
                : ParsedStatement(start, length), _condition(condition), _block(block), _elseStatement(nullptr) {
        }

        ParsedConditionalStatement(ParsedExpression *condition, ParsedStatement *block, ParsedStatement *nextStatement,
                                   unsigned int start, unsigned int length)
                : ParsedStatement(start, length), _condition(condition), _block(block), _elseStatement(nextStatement) {
            _condition = condition;
            _block = block;
            _elseStatement = nextStatement;
        }

        ~ParsedConditionalStatement() final {
            delete _condition;
            delete _block;
            delete _elseStatement;
        }

        const ParsedStatementKind GetKind() const final {
            return ParsedStatementKind::Conditional;
        }

        const ParsedExpression *GetCondition() const {
            return _condition;
        }

        const ParsedStatement *GetBlock() const {
            return _block;
        }

        const ParsedStatement *GetElseStatement() const {
            return _elseStatement;
        }
    };
}
#endif //PORYGONLANG_PARSEDSTATEMENT_HPP
